@@ -1,5 +1,8 @@
 """Some utilities for manipulating GeoSpatial data."""
+import contextlib
 import numbers
+import os
+from pathlib import Path
 from typing import Any, Dict, List, Optional, Tuple, Union
 from warnings import warn
 
@@ -48,14 +51,23 @@ def json2geodf(
     content = content if isinstance(content, list) else [content]
     try:
         geodf = gpd.GeoDataFrame.from_features(content[0], crs=in_crs)
+    except AttributeError:
+        geodf = gpd.GeoDataFrame.from_features(content[0])
     except TypeError:
         content = [arcgis2geojson(c) for c in content]
-        geodf = gpd.GeoDataFrame.from_features(content[0], crs=in_crs)
+        try:
+            geodf = gpd.GeoDataFrame.from_features(content[0], crs=in_crs)
+        except AttributeError:
+            geodf = gpd.GeoDataFrame.from_features(content[0])
 
     if len(content) > 1:
-        geodf = geodf.append([gpd.GeoDataFrame.from_features(c, crs=in_crs) for c in content[1:]])
+        for c in content[1:]:
+            try:
+                geodf = geodf.append(gpd.GeoDataFrame.from_features(c, crs=in_crs))
+            except AttributeError:
+                geodf = gpd.GeoDataFrame.from_features(c)
 
-    if in_crs != crs:
+    if in_crs != crs and "geometry" in geodf:
         geodf = geodf.to_crs(crs)
 
     return geodf
@@ -81,9 +93,9 @@ def arcgis2geojson(arcgis: Dict[str, Any], id_attr: Optional[str] = None) -> Dic
         A GeoJSON file readable by GeoPandas
     """
 
-    def convert(arcgis, id_attr=None):
+    def convert(arcgis: Dict[str, Any], id_attr: Optional[str] = None) -> Dict[str, Any]:
         """Convert an ArcGIS JSON object to a GeoJSON object."""
-        geojson = {}
+        geojson: Dict[str, Any] = {}
 
         if "features" in arcgis and arcgis["features"]:
             geojson["type"] = "FeatureCollection"
@@ -164,7 +176,7 @@ def arcgis2geojson(arcgis: Dict[str, Any], id_attr: Optional[str] = None) -> Dic
 
         return geojson
 
-    def rings2geojson(rings):
+    def rings2geojson(rings: List[List[List[float]]]) -> Dict[str, Any]:
         """Check for holes in the ring and fill them."""
         outer_rings = []
         holes = []
@@ -251,6 +263,36 @@ def arcgis2geojson(arcgis: Dict[str, Any], id_attr: Optional[str] = None) -> Dic
     return convert(arcgis, id_attr)
 
 
+def _get_nodata_crs(resp: bytes, driver: str) -> Tuple[np.float64, pyproj.crs.crs.CRS]:
+    """Get nodata and crs value of a raster in bytes.
+
+    Parameters
+    ----------
+    resp : bytes
+        Raster response returned from a wed service request such as WMS
+    driver : str
+        A GDAL driver for reading the content, defaults to GTiff. A list of the drivers
+        can be found here: https://gdal.org/drivers/raster/index.html
+
+    Returns
+    -------
+    tuple
+        (NoData, CRS)
+    """
+    with rio.MemoryFile() as memfile:
+        memfile.write(resp)
+        with memfile.open(driver=driver) as src:
+            r_crs = pyproj.CRS.from_user_input(src.crs)
+            if src.nodata is None:
+                try:
+                    nodata = np.iinfo(src.dtypes[0]).max
+                except ValueError:
+                    nodata = np.nan
+            else:
+                nodata = np.dtype(src.dtypes[0]).type(src.nodata)
+    return nodata, r_crs
+
+
 def gtiff2xarray(
     r_dict: Dict[str, bytes],
     geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
@@ -286,33 +328,17 @@ def gtiff2xarray(
     key1 = next(iter(r_dict.keys()))
     if len(r_dict) == 1 and "dd" not in key1:
         r_dict = {f"{key1}_dd_0_0": r_dict[key1]}
+        key1 = f"{key1}_dd_0_0"
 
     var_name = {lyr: "_".join(lyr.split("_")[:-3]) for lyr in r_dict.keys()}
 
-    with rio.MemoryFile() as memfile:
-        memfile.write(r_dict[next(iter(r_dict.keys()))])
-        with memfile.open(driver=driver) as src:
-            r_crs = pyproj.CRS.from_user_input(src.crs)
-            if src.nodata is None:
-                try:
-                    nodata = np.iinfo(src.dtypes[0]).max
-                except ValueError:
-                    nodata = np.nan
-            else:
-                nodata = np.dtype(src.dtypes[0]).type(src.nodata)
-
-            ds = xr.open_rasterio(src)
-            valid_dims = list(ds.sizes)
-            if any(d not in valid_dims for d in ds_dims):
-                raise InvalidInputValue("ds_dims", valid_dims)
+    nodata, r_crs = _get_nodata_crs(r_dict[key1], driver)
 
     _geometry = geo2polygon(geometry, geo_crs, r_crs)
-    if not _geometry.is_valid:
-        _geometry = _geometry.buffer(0.0)
 
-    def to_dataset(lyr: str) -> xr.DataArray:
+    def to_dataset(lyr: str, resp: bytes) -> xr.DataArray:
         with rio.MemoryFile() as memfile:
-            memfile.write(r_dict[lyr])
+            memfile.write(resp)
             with memfile.open(driver=driver) as src:
                 geom = [_geometry.intersection(box(*src.bounds))]
                 if geom[0].is_empty:
@@ -332,10 +358,12 @@ def gtiff2xarray(
 
                 with rio.vrt.WarpedVRT(src, **meta) as vrt:
                     ds = xr.open_rasterio(vrt)
-                    try:
+                    valid_dims = list(ds.sizes)
+                    if any(d not in valid_dims for d in ds_dims):
+                        raise InvalidInputValue("ds_dims", valid_dims)
+                    with contextlib.suppress(ValueError):
                         ds = ds.squeeze("band", drop=True)
-                    except ValueError:
-                        pass
+
                     coords = {ds_dims[0]: ds.coords[ds_dims[0]], ds_dims[1]: ds.coords[ds_dims[1]]}
                     msk_da = xr.DataArray(msk, coords, dims=ds_dims)
                     ds = ds.where(msk_da, drop=True)
@@ -343,7 +371,7 @@ def gtiff2xarray(
                     ds.name = var_name[lyr]
                     return ds
 
-    ds = xr.merge(to_dataset(lyr) for lyr in r_dict.keys())
+    ds = xr.merge(to_dataset(lyr, resp) for lyr, resp in r_dict.items())
     ds.attrs["crs"] = r_crs.to_string()
 
     if len(ds.variables) - len(ds.dims) == 1:
@@ -351,12 +379,74 @@ def gtiff2xarray(
     return ds
 
 
+def gtiff2file(
+    r_dict: Dict[str, bytes],
+    geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
+    geo_crs: str,
+    output: Union[str, Path] = ".",
+    driver: str = "GTiff",
+) -> None:
+    """Save responses from ``pygeoogc.wms_bybox`` to raster file(s).
+
+    Parameters
+    ----------
+    r_dict : dict
+        The output of ``wms_bybox`` function.
+    geometry : Polygon, MultiPolygon, or tuple
+        The geometry to mask the data that should be in the same CRS as the r_dict.
+    geo_crs : str
+        The spatial reference of the input geometry.
+    output : str
+        Path to a folder saving files. File names are keys of the input dictionary, so
+        each layer becomes one file. Defaults to current directory.
+    driver : str, optional
+        A GDAL driver for reading the content, defaults to GTiff. A list of the drivers
+        can be found here: https://gdal.org/drivers/raster/index.html
+    """
+    os.makedirs(output, exist_ok=True)
+
+    if not isinstance(r_dict, dict):
+        raise InvalidInputType("r_dict", "dict", '{"name": Response.content}')
+
+    key1 = next(iter(r_dict.keys()))
+    if len(r_dict) == 1 and "dd" not in key1:
+        r_dict = {f"{key1}_dd_0_0": r_dict[key1]}
+        key1 = f"{key1}_dd_0_0"
+
+    nodata, r_crs = _get_nodata_crs(r_dict[key1], driver)
+
+    _geometry = geo2polygon(geometry, geo_crs, r_crs)
+
+    for lyr, resp in r_dict.items():
+        with rio.MemoryFile() as memfile:
+            memfile.write(resp)
+            with memfile.open(driver=driver) as src:
+                geom = [_geometry.intersection(box(*src.bounds))]
+                if geom[0].is_empty:
+                    data, transform = rio_mask.mask(src, [_geometry], crop=True)
+                else:
+                    data, transform = rio_mask.mask(src, geom, crop=True)
+                meta = src.meta
+                meta.update(
+                    {
+                        "driver": "GTiff",
+                        "height": data.shape[1],
+                        "width": data.shape[2],
+                        "transform": transform,
+                        "nodata": nodata,
+                    }
+                )
+
+                with rio.open(Path(output, f"{lyr}.gtiff"), "w", **meta) as dest:
+                    dest.write(data)
+
+
 def xarray_geomask(
     ds: Union[xr.Dataset, xr.DataArray],
     geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
     geo_crs: str,
     ds_dims: Tuple[str, str] = ("y", "x"),
-):
+) -> Union[xr.Dataset, xr.DataArray]:
     """Mask a ``xarray.Dataset`` based on a geometry.
 
     Parameters
@@ -465,7 +555,7 @@ class MatchCRS:
             raise InvalidInputType("geom", "tuple of length 4", "(west, south, east, north)")
 
         project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform
-        return ops.transform(project, box(*geom)).bounds
+        return ops.transform(project, box(*geom)).bounds  # type: ignore
 
     @staticmethod
     def coords(
@@ -506,10 +596,14 @@ def geo2polygon(
     Polygon
         A Polygon in the target CRS.
     """
+    if not isinstance(geometry, (Polygon, MultiPolygon, tuple)):
+        raise InvalidInputType("geometry", "Polygon, MultiPolygon, or tuple of length 4")
+
     if isinstance(geometry, tuple):
-        return box(*MatchCRS.bounds(geometry, geo_crs, crs))  # type: ignore
+        geom = box(*MatchCRS.bounds(geometry, geo_crs, crs))  # type: ignore
+    elif isinstance(geometry, (Polygon, MultiPolygon)):
+        geom = MatchCRS.geometry(geometry, geo_crs, crs)
 
-    if isinstance(geometry, (Polygon, MultiPolygon)):
-        return MatchCRS.geometry(geometry, geo_crs, crs)
-
-    raise InvalidInputType("geometry", "Polygon, MultiPolygon, or tuple of length 4")
+    if not geom.is_valid:
+        geom = geom.buffer(0.0)
+    return geom
